package com.mx.finger.dome;

import android.Manifest;
import android.app.ActionBar;
import android.content.BroadcastReceiver;
import android.content.Context;
import android.content.Intent;
import android.content.IntentFilter;
import android.content.pm.PackageManager;
import android.graphics.Bitmap;
import android.os.Build;
import android.os.Bundle;
import android.support.annotation.Nullable;
import android.support.v7.app.AppCompatActivity;
import android.text.Editable;
import android.util.Log;
import android.view.View;
import android.widget.EditText;
import android.widget.ImageView;
import android.widget.TextView;
import android.zyapi.CommonApi;

import com.mx.finger.alg.MxIDFingerAlg;
import com.mx.finger.api.msc.MxIdMscFingerApiFactory;
import com.mx.finger.api.msc.MxMscBigFingerApi;
import com.mx.finger.common.MxImage;
import com.mx.finger.common.Result;
import com.mx.finger.utils.LogUtils;
import com.mx.finger.utils.RawBitmapUtils;

import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;

public class MainActivity extends AppCompatActivity {

    private static final String TAG = "Mx-MainActivity";
    private static final int TIME_OUT = 5000;
    private final ExecutorService executor = Executors.newSingleThreadExecutor();

    private byte[] featureBufferEnroll = new byte[1];
    private byte[] featureBufferMatch = new byte[1];

    private MxMscBigFingerApi mxMscBigFingerApi;
    private MxIDFingerAlg mxFingerAlg;
    private SearchThread mSearchThread;

    private static int mComFd = -1;
    static CommonApi mCommonApi;

    @Override
    protected void onCreate(@Nullable Bundle savedInstanceState) {
        super.onCreate(savedInstanceState);
        setContentView(R.layout.activity_main2);

        openGPIO();

        // show logs
        ActionBar actionBar = getActionBar();
        if (actionBar != null) {
            actionBar.setTitle(R.string.app_name);
        }

        if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.M) {
            if (checkSelfPermission(Manifest.permission.READ_PHONE_STATE) != PackageManager.PERMISSION_GRANTED
                    || checkSelfPermission(Manifest.permission.READ_EXTERNAL_STORAGE) != PackageManager.PERMISSION_GRANTED) {
                requestPermissions(new String[]{Manifest.permission.READ_EXTERNAL_STORAGE, Manifest.permission.READ_PHONE_STATE}, 0x99);
            }
        }

        LogUtils.logAble(true);
        MxIdMscFingerApiFactory fingerFactory = new MxIdMscFingerApiFactory(getApplication());
        mxMscBigFingerApi = fingerFactory.getApi();
        mxFingerAlg = fingerFactory.getAlg();


        findViewById(R.id.getVersion).setOnClickListener(v -> executor.execute(() -> {
            String version = mxFingerAlg.version();
            showLog("GetVersion mxFingerAlg success \n" + version);
            Result<String> deviceInfo = mxMscBigFingerApi.getDeviceInfo();
            if (deviceInfo.isSuccess()) {
                showLog("GetVersion success \n" + deviceInfo.data);
            } else {
                showLog("GetVersion fail \nCode  : " + deviceInfo.code);
            }
        }));

        findViewById(R.id.getSerialNumber).setOnClickListener(v -> executor.execute(() -> {
            Result<String> result = mxMscBigFingerApi.getDeviceSerialNumber();
            if (result.isSuccess()) {
                showLog("GetSerialNumber success \n" + result.data);
            } else {
                showLog("GetSerialNumber fail \nCode  : " + result.code);
            }
        }));

        findViewById(R.id.detectFinger).setOnClickListener(v -> executor.execute(() -> {
            Result<Boolean> result = mxMscBigFingerApi.detectFinger();
            if (result.isSuccess()) {
                showLog("DetectFinger success \n" + (result.data != null && result.data ? " Has Finger" : "No Finger"));
            } else {
                showLog("DetectFinger fail \nCode  : " + result.code);
            }
        }));

        findViewById(R.id.getFingerImage).setOnClickListener(v -> executor.execute(() -> {
            showProgressDialog("Please press finger ");
            Result<MxImage> result = mxMscBigFingerApi.getFingerImageBig(TIME_OUT);
            if (result.isSuccess()) {
                showLog("GetFingerImage success !");
                showFingerImage(result.data);
            } else {
                showLog("GetFingerImage fail \nCode  : " + result.code);
                showFingerImage(null);
            }
            dismissProgressDialog();
        }));


        findViewById(R.id.enrollFingerHost).setOnClickListener(v -> executor.execute(() -> {
            showProgressDialog("Please press finger ");
            try {
                //step 0 get finger image
                Result<MxImage> result = mxMscBigFingerApi.getFingerImageBig(TIME_OUT);
                if (!result.isSuccess()) {
                    showLog("GetFingerImage fail \nCode  : " + result.code);
                    showFingerImage(null);
                    return;
                }
                //step 1 extract finger feature
                MxImage image = result.data;
                showFingerImage(image);
                assert image != null;
                featureBufferEnroll = mxFingerAlg.extractFeature(image.data, image.width, image.height);
                if (featureBufferEnroll == null) {
                    appendLog("EnrollFingerHost fail ");
                } else {
                    appendLog("EnrollFingerHost success ");
                }
            } finally {
                dismissProgressDialog();
            }
        }));

        findViewById(R.id.matchFingerHost).setOnClickListener(v -> executor.execute(() -> {
            showProgressDialog("Please press finger ");
            try {
                //step 0 get finger image
                Result<MxImage> result = mxMscBigFingerApi.getFingerImageBig(TIME_OUT);
                if (!result.isSuccess()) {
                    showLog("GetFingerImage fail \nCode : " + result.code);
                    showFingerImage(null);
                    return;
                }
                MxImage image = result.data;
                showFingerImage(image);
                //step 1 get finger feature
                assert image != null;
                featureBufferMatch = mxFingerAlg.extractFeature(image.data, image.width, image.height);
                if (featureBufferMatch == null) {
                    appendLog("ExtractFeature fail");
                    return;
                }
                //step 2 match finger feature
                int match = mxFingerAlg.match(featureBufferEnroll, featureBufferMatch, 3);
                if (match == 0) {
                    appendLog("MatchFingerFeature success");
                } else {
                    appendLog("MatchFingerFeature fail \nCode : " + match);
                }
            } finally {
                dismissProgressDialog();
            }
        }));

        findViewById(R.id.flashSize).setOnClickListener(v -> executor.execute(() -> {
            Result<int[]> result = mxMscBigFingerApi.getFlashSize();
            if (result.isSuccess()) {
                assert result.data != null;
                showLog(String.format("GetFlashSize success  \nRemaining size: %s \nAll size: %s  ", result.data[1], result.data[0]));
            } else {
                showLog("GetFlashSize fail \nCode : " + result.code);
            }
        }));

        findViewById(R.id.enrollTemplate).setOnClickListener(v -> executor.execute(() -> {
            showFingerImage(null);
            showProgressDialog("Please press finger");
            try {
                int position = getTemplatePosition();
                if (position < 0) {
                    return;
                }
                // if you wan't test download finger feature, use this line .
                // Result<Boolean> saveResult = enrollTemplateWithHostFeature(position);
                Result<Boolean> saveResult = mxMscBigFingerApi.extractFingerFeatureBig(position, MxMscBigFingerApi.FLASH, TIME_OUT);
                if (!saveResult.isSuccess()) {
                    showLog("EnrollTemplate failed \nCode : " + saveResult.code);
                } else {
                    showLog("EnrollTemplate success\nPosition : " + position);
                }
            } finally {
                dismissProgressDialog();
            }
        }));

        findViewById(R.id.matchTemplate).setOnClickListener(v -> executor.execute(() -> {
            showFingerImage(null);
            showProgressDialog("Please press finger ");
            try {
                int position = getTemplatePosition();
                if (position < 0) {
                    return;
                }
                Result<Boolean> result = mxMscBigFingerApi.extractFingerFeatureBig(0, MxMscBigFingerApi.RAM, TIME_OUT);
                if (!result.isSuccess()) {
                    showLog("ExtractFeature \nCode  : " + result.code);
                    return;
                }
                Result<Boolean> booleanResult = mxMscBigFingerApi.match(0, MxMscBigFingerApi.RAM, position, MxMscBigFingerApi.FLASH);
                if (booleanResult.isSuccess()) {
                    showLog("MatchFinger success ! ");
                } else {
                    showLog("MatchFinger fail \nCode : " + booleanResult.code);
                }
            } finally {
                dismissProgressDialog();
            }
        }));

        findViewById(R.id.deleteTemplate).setOnClickListener(v -> executor.execute(() -> {
            showFingerImage(null);
            int templateIndex = getTemplatePosition();
            if (templateIndex < 0) {
                return;
            }
            Result<Boolean> deleteResult = mxMscBigFingerApi.deleteFlashTemplate(templateIndex);
            if (!deleteResult.isSuccess()) {
                showLog("DeleteTemplate failed \nCode : " + deleteResult.code);
            } else {
                showLog("DeleteTemplate success!");
            }
        }));

        findViewById(R.id.clearTemplate).setOnClickListener(v -> executor.execute(() -> {
            showFingerImage(null);
            Result<Boolean> clearResult = mxMscBigFingerApi.clearFlashTemplate();
            if (!clearResult.isSuccess()) {
                showLog("ClearTemplate failed \nCode : " + clearResult.code);
            } else {
                showLog("ClearTemplate success");
            }
        }));

        findViewById(R.id.searchTemplate).setOnClickListener(v -> {
            if (mSearchThread == null) {
                // start search
                mSearchThread = new SearchThread();
                ((TextView) v).setText(R.string.stop_search);
                setButtonsEnable(false);
                mSearchThread.start();
            } else {
                // stop search
                mSearchThread.interrupt();
                try {
                    mSearchThread.join();
                } catch (InterruptedException e) {
                    e.printStackTrace();
                }
                mSearchThread = null;
                ((TextView) v).setText(R.string.search);
                setButtonsEnable(true);
            }
        });

        greenOnReceiver = new GreenOnReceiver();

        IntentFilter filter = new IntentFilter();
        filter.addAction(Intent.ACTION_SCREEN_ON);
        filter.addAction(Intent.ACTION_SCREEN_OFF);
        filter.addAction(Intent.ACTION_USER_PRESENT);
        registerReceiver(greenOnReceiver, filter);

    }

    GreenOnReceiver greenOnReceiver;
    class GreenOnReceiver extends BroadcastReceiver {
        @Override
        public void onReceive(Context context, Intent intent) {
            // TODO Auto-generated method stub

            if (intent.getAction().equals(Intent.ACTION_SCREEN_OFF)) {
                closeGPIO();

            } else if (intent.getAction().equals(Intent.ACTION_SCREEN_ON)) {

                openGPIO();
            }
        }
    }

    class SearchThread extends Thread {
        @Override
        public void run() {
            super.run();
            int index = 0;
            while (!isInterrupted()) {
                // if you don't need show finger image , use this line .
                //Result<Integer> searchResult = mxMscBigFingerApi.findFingerTemplate();
                Result<Integer> searchResult = findTemplateWithImage();
                if (searchResult.isSuccess()) {
                    showLog("Count " + index++ + "\nSearch success !\nPosition :" + searchResult.data);
                } else {
                    String reason;
                    switch (searchResult.code) {
                        case MxMscBigFingerApi.CODE_NO_FINGER:
                            reason = "No Finger !";
                            break;
                        case MxMscBigFingerApi.CODE_FAIL:
                            reason = "Not Found !";
                            break;
                        default:
                            reason = "Error !";
                    }
                    showLog("Count " + index++ + "\nSearch failed  !\nReason  : " + reason);
                }
            }
        }
    }

    Result<Integer> findTemplateWithImage() {
        // detect finger
        Result<Boolean> detectFinger = mxMscBigFingerApi.detectFinger();
        if (!detectFinger.isSuccess()) {
            return Result.error(detectFinger.code);
        }
        if (detectFinger.data == null || !detectFinger.data) {
            return Result.error(MxMscBigFingerApi.CODE_NO_FINGER);
        }
        // read finger image to image cache
        Result<Boolean> readFingerImage = mxMscBigFingerApi.readFingerImage();
        if (!readFingerImage.isSuccess()) {
            return Result.error(readFingerImage.code);
        }
        // upload cache image to host
        Result<MxImage> fingerImageBig = mxMscBigFingerApi.uploadFingerImageBig();
        if (!fingerImageBig.isSuccess()) {
            return Result.error(fingerImageBig.code);
        }
        showFingerImage(fingerImageBig.data);
        // read finger feature to ram 0
        final int RAM_INDEX = 0;
        Result<Boolean> readFingerFeatureBig = mxMscBigFingerApi.readFingerFeatureBig(RAM_INDEX);
        if (!readFingerFeatureBig.isSuccess()) {
            return Result.error(readFingerFeatureBig.code);
        }
        return mxMscBigFingerApi.searchFlashTemplate(RAM_INDEX);
    }

    Result<Boolean> enrollTemplateWithHostFeature(int position) {
        //step 0 get finger image
        Result<MxImage> result = mxMscBigFingerApi.getFingerImageBig(TIME_OUT);
        if (!result.isSuccess()) {
            showFingerImage(null);
            return Result.error(result.code);
        }
        MxImage image = result.data;
        showFingerImage(image);
        //step 1 get finger feature
        assert image != null;
        byte[] featureResult = mxFingerAlg.extractFeature(image.data, image.width, image.height);
        if (featureResult == null) {
            return Result.error(999);
        }
        //step 2 down finger feature
        return mxMscBigFingerApi.downloadFlashTemplate(position, featureResult);
    }

    int getTemplatePosition() {
        EditText editText = findViewById(R.id.et_template);
        Editable text = editText.getText();
        try {
            int index = Integer.parseInt(text.toString());
            if (index > 127 || index < 0) {
                showLog("template position is beyond limits");
                return index;
            }
            return index;
        } catch (Exception e) {
            showLog("please input template position");
            return -1;
        }
    }

    void showProgressDialog(String msg) {
        runOnUiThread(() -> {
            showLog(msg);
            setButtonsEnable(false);
            findViewById(R.id.progress_horizontal).setVisibility(View.VISIBLE);
        });
    }

    void dismissProgressDialog() {
        runOnUiThread(() -> {
            setButtonsEnable(true);
            findViewById(R.id.progress_horizontal).setVisibility(View.GONE);
        });
    }

    void setButtonsEnable(boolean enable) {
        findViewById(R.id.getVersion).setEnabled(enable);
        findViewById(R.id.getSerialNumber).setEnabled(enable);
        findViewById(R.id.detectFinger).setEnabled(enable);
        findViewById(R.id.getFingerImage).setEnabled(enable);
        findViewById(R.id.enrollFingerHost).setEnabled(enable);
        findViewById(R.id.matchFingerHost).setEnabled(enable);
        findViewById(R.id.flashSize).setEnabled(enable);
        findViewById(R.id.matchTemplate).setEnabled(enable);
        findViewById(R.id.et_template).setEnabled(enable);
        findViewById(R.id.enrollTemplate).setEnabled(enable);
        findViewById(R.id.deleteTemplate).setEnabled(enable);
        findViewById(R.id.clearTemplate).setEnabled(enable);
        findViewById(R.id.searchTemplate).setEnabled(enable || mSearchThread != null);
    }

    void showFingerImage(MxImage image) {
        runOnUiThread(() -> {
            Bitmap bitmap = image == null ? null : RawBitmapUtils.raw2Bimap(image.data, image.width, image.height);
            ImageView imageView = findViewById(R.id.image);
            if (bitmap == null) {
                imageView.setVisibility(View.GONE);
            } else {
                imageView.setVisibility(View.VISIBLE);
                imageView.setImageBitmap(bitmap);
            }
        });
    }

    void appendLog(String msg) {
        Log.i(TAG, "appendLog: " + msg);
        runOnUiThread(() -> {
            TextView textView = findViewById(R.id.log);
            textView.append(msg);
            textView.append("\n");
        });
    }

    void showLog(String msg) {
        Log.i(TAG, "showLog: " + msg);
        runOnUiThread(() -> {
            TextView textView = findViewById(R.id.log);
            textView.setText(msg);
            textView.append("\n");
        });
    }

    @Override
    protected void onDestroy() {
        closeGPIO();
        unregisterReceiver(greenOnReceiver);
        super.onDestroy();
    }

    /**
     * 打开GPIO口，把电平转移到指纹模块
     */
    private void openGPIO() {

        mCommonApi = new CommonApi();

        mCommonApi.setGpioDir(18, 1);
        mCommonApi.setGpioOut(18, 1);

        try {
            Thread.sleep(200);
        } catch (InterruptedException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        }

        mCommonApi.setGpioDir(107, 1);
        mCommonApi.setGpioOut(107, 1);

//        mCommonApi.setGpioDir(25, 1);
//        mCommonApi.setGpioOut(25, 1);
//
//        try {
//            Thread.sleep(200);
//        } catch (InterruptedException e) {
//            // TODO Auto-generated catch block
//            e.printStackTrace();
//        }
//
//        mCommonApi.setGpioDir(26, 1);
//        mCommonApi.setGpioOut(26, 1);
//
//        try {
//            Thread.sleep(200);
//        } catch (InterruptedException e) {
//            // TODO Auto-generated catch block
//            e.printStackTrace();
//        }
//
//        mCommonApi.setGpioDir(19, 1);
//        mCommonApi.setGpioOut(19, 1);
    }

    /**
     * 关闭GPIO口，把电平转移到USB
     */
    private void closeGPIO(){

        if(mCommonApi==null){

            mCommonApi=new CommonApi();

        }

        mCommonApi.setGpioDir(18, 1);
        mCommonApi.setGpioOut(18, 0);

        try {
            Thread.sleep(200);
        } catch (InterruptedException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        }

        mCommonApi.setGpioDir(107, 1);
        mCommonApi.setGpioOut(107, 0);

//        mCommonApi.setGpioDir(19, 1);
//        mCommonApi.setGpioOut(19, 0);
//
//        mCommonApi.setGpioDir(25, 1);
//        mCommonApi.setGpioOut(25, 0);
//
//        mCommonApi.setGpioDir(26, 1);
//        mCommonApi.setGpioOut(26, 0);

    }
}
